/*
 * Test file removal semantics.
 *
 * Copyright (C) 2019, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <ext2fs/ext2fs.h>

#include "file.h"
#include "format.h"
#include "image.h"
#include "path.h"
#include "utils.h"



/* Buffer size for file access. */

const int BUFSIZE = 4096;

/* Number of open files. */

const int FILES = 20;



int main(int argc, char *argv[])
{
    int flags = EXT2_FLAG_RW;
    char *fsname;
    ext2_filsys fs = NULL;
    errcode_t retval;
    int exitcode = 0;
    char basename[32];
    ext2_ino_t ino, inos[FILES];
    ext2_file_t file, files[FILES];
    char buf[BUFSIZE];
    unsigned int transferred;
    int i;

    if (argc < 2)
    {
        printf("Usage: %s <device or image file>\n", argv[0]);
        return 1;
    }

    fsname = argv[1];

    /* Open the filesystem image using the POSIX file access mechanism. */

    retval = ext2fs_open(fsname, flags, 0, 0, unix_io_manager, &fs);
    if (retval)
    {
        printf("Could not open filesystem using %s\n", fsname);
        return 1;
    }

    retval = ext2fs_read_bitmaps(fs);
    if (retval)
    {
        printf("Could not read bitmaps from %s\n", fsname);
        return 1;
    }

    printf("Filesystem with blocksize %d\n", fs->blocksize);

    /* Create some files. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i);

        retval = image_create_file(fs, EXT2_ROOT_INO, basename, 0644,
                                   1000, 1000, &ino);

        if (retval)
        {
            printf("Could not create file %s\n", basename);
            return 1;
        }

        if (ext2fs_file_open(fs, ino, EXT2_FILE_WRITE | EXT2_FILE_CREATE, &file))
        {
            printf("Could not open file %s\n", basename);
            return 1;
        }

        sprintf(buf, "writing to file%04d", i);

        if (ext2fs_file_write(file, buf, strlen(buf), &transferred))
        {
            printf("Could not write to file %s\n", basename);
            return 1;
        }

        ext2fs_file_flush(file);

        files[i - 1] = file;
        inos[i - 1] = ino;
    }

    utils_list_dir(fs, "");
    printf("----\n");

    /* Unlink the files. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i);

        if (ext2fs_unlink(fs, EXT2_ROOT_INO, basename, 0, 0))
        {
            printf("Could not unlink file %s\n", basename);
            return 1;
        }
    }

    utils_list_dir(fs, "");
    printf("----\n");

    /* Access and close unlinked files. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i);
        sprintf(buf, "; writing to file%04d", i);

        file = files[i - 1];

        if (ext2fs_file_write(file, buf, strlen(buf), &transferred))
        {
            printf("Could not write to file %s\n", basename);
            return 1;
        }

        ext2fs_file_flush(file);
        ext2fs_file_close(file);
    }

    utils_list_dir(fs, "");
    printf("----\n");

    /* Create some more files. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i + FILES);

        retval = image_create_file(fs, EXT2_ROOT_INO, basename, 0644,
                                   1000, 1000, &ino);

        if (retval)
        {
            printf("Could not create file %s\n", basename);
            return 1;
        }

        if (ext2fs_file_open(fs, ino, EXT2_FILE_WRITE | EXT2_FILE_CREATE, &file))
        {
            printf("Could not write file %s\n", basename);
            return 1;
        }

        ext2fs_file_flush(file);
        ext2fs_file_close(file);
    }

    utils_list_dir(fs, "");
    printf("----\n");

    /* Re-open the original files and read from them. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i);

        ino = inos[i - 1];

        if (ext2fs_file_open(fs, ino, 0, &file))
        {
            printf("Could not open file %s\n", basename);
            return 1;
        }

        if (ext2fs_file_read(file, buf, BUFSIZE, &transferred))
        {
            printf("Could not read from file %s\n", basename);
            return 1;
        }

        buf[transferred] = '\0';
        printf("Read from %s: %s\n", basename, buf);

        ext2fs_file_close(file);
    }

    /* Create some more files with the same names as the original ones. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i);

        retval = image_create_file(fs, EXT2_ROOT_INO, basename, 0644,
                                   1000, 1000, &ino);

        if (retval)
        {
            printf("Could not create file %s\n", basename);
            return 1;
        }

        if (ext2fs_file_open(fs, ino, EXT2_FILE_WRITE | EXT2_FILE_CREATE, &file))
        {
            printf("Could not write file %s\n", basename);
            return 1;
        }

        ext2fs_file_flush(file);
        ext2fs_file_close(file);
    }

    utils_list_dir(fs, "");
    printf("----\n");

    /* Re-open the original files and read from them again. */

    for (i = 1; i <= FILES; i++)
    {
        sprintf(basename, "file%04d", i);

        ino = inos[i - 1];

        if (ext2fs_file_open(fs, ino, 0, &file))
        {
            printf("Could not open file %s\n", basename);
            return 1;
        }

        if (ext2fs_file_read(file, buf, BUFSIZE, &transferred))
        {
            printf("Could not read from file %s\n", basename);
            return 1;
        }

        buf[transferred] = '\0';
        printf("Read from %s: %s\n", basename, buf);

        ext2fs_file_close(file);
    }

    /* Close the filesystem image. */

    retval = ext2fs_flush(fs);
    if (retval)
    {
        printf("Error flushing filesystem in %s\n", fsname);
        exitcode = 1;
    }

    retval = ext2fs_close(fs);
    if (retval)
    {
        printf("Error closing filesystem in %s\n", fsname);
        exitcode = 1;
    }

    ext2fs_free(fs);
    return exitcode;
}
